from typing import Literal

SupervisorTaskStep = Literal[
    "configure-supervisor",
    "start-supervisor",
    "start-programs",
    "stop-programs",
    "check-programs",
    "restart-programs",
    "stop-supervisor",
    "unconfigure-supervisor",
    "force-kill",
]

__all__ = ("SupervisorTaskStep",)



from datetime import datetime

from supervisor_pydantic.client import ProcessInfo, ProcessState


def _gen() -> ProcessInfo:
    return ProcessInfo(
        name="test",
        group="test",
        state=ProcessState.UNKNOWN,
        description="",
        start=datetime.now(),
        stop=datetime.now(),
        now=datetime.now(),
        spawner="",
        exitstatus=0,
        logfile="",
        stdout_logfile="",
        stderr_logfile="",
        pid=0,
    )


def test_ok():
    x = _gen()
    x.state = ProcessState.RUNNING
    assert x.ok()
    x.state = ProcessState.EXITED
    x.exitstatus = 0
    assert x.ok()



from supervisor_pydantic import RpcInterfaceConfiguration


def test_inst():
    RpcInterfaceConfiguration()
    RpcInterfaceConfiguration(rpcinterface_factory="a.test.module")


def test_cfg():
    c = RpcInterfaceConfiguration(rpcinterface_factory="a.test.module")
    assert c.to_cfg("name").strip() == "[rpcinterface:name]\nsupervisor.rpcinterface_factory=a.test.module"



from pathlib import Path
from typing import Optional

from pydantic import Field, field_serializer, field_validator

from .base import LogLevel, OctalUmask, UnixUserName, _BaseCfgModel

__all__ = ("SupervisordConfiguration",)


class SupervisordConfiguration(_BaseCfgModel):
    logfile: Optional[Path] = Field(
        default=None,
        description="The path to the activity log of the supervisord process. This option can include the value %(here)s, which expands to the directory in which the supervisord configuration file was found.",
    )
    logfile_maxbytes: Optional[str] = Field(
        default=None,
        description="The maximum number of bytes that may be consumed by the activity log file before it is rotated (suffix multipliers like “KB”, “MB”, and “GB” can be used in the value). Set this value to 0 to indicate an unlimited log size.",
    )
    logfile_backups: Optional[int] = Field(
        default=None,
        description="The number of backups to keep around resulting from activity log file rotation. If set to 0, no backups will be kept.",
    )
    loglevel: Optional[LogLevel] = Field(
        default=None,
        description="The logging level, dictating what is written to the supervisord activity log. One of critical, error, warn, info, debug, trace, or blather. Note that at log level debug, the supervisord log file will record the stderr/stdout output of its child processes and extended info about process state changes, which is useful for debugging a process which isn’t starting properly. See also: Activity Log Levels.",
    )
    pidfile: Optional[Path] = Field(
        default=None,
        description="The location in which supervisord keeps its pid file. This option can include the value %(here)s, which expands to the directory in which the supervisord configuration file was found.",
    )
    umask: Optional[OctalUmask] = Field(default=None, description="The umask of the supervisord process.")
    nodaemon: Optional[bool] = Field(default=None, description="If true, supervisord will start in the foreground instead of daemonizing.")
    silent: Optional[bool] = Field(default=None, description="If true and not daemonized, logs will not be directed to stdout.")
    minfds: Optional[int] = Field(
        default=None,
        description="The minimum number of file descriptors that must be available before supervisord will start successfully. A call to setrlimit will be made to attempt to raise the soft and hard limits of the supervisord process to satisfy minfds. The hard limit may only be raised if supervisord is run as root. supervisord uses file descriptors liberally, and will enter a failure mode when one cannot be obtained from the OS, so it’s useful to be able to specify a minimum value to ensure it doesn’t run out of them during execution. These limits will be inherited by the managed subprocesses. This option is particularly useful on Solaris, which has a low per-process fd limit by default.",
    )
    minprocs: Optional[int] = Field(
        default=None,
        description="The minimum number of process descriptors that must be available before supervisord will start successfully. A call to setrlimit will be made to attempt to raise the soft and hard limits of the supervisord process to satisfy minprocs. The hard limit may only be raised if supervisord is run as root. supervisord will enter a failure mode when the OS runs out of process descriptors, so it’s useful to ensure that enough process descriptors are available upon supervisord startup.",
    )
    nocleanup: Optional[bool] = Field(
        default=None,
        description="Prevent supervisord from clearing any existing AUTO child log files at startup time. Useful for debugging.",
    )
    childlogdir: Optional[Path] = Field(
        default=None,
        description="The directory used for AUTO child log files. This option can include the value %(here)s, which expands to the directory in which the supervisord configuration file was found.",
    )
    user: Optional[UnixUserName] = Field(
        default=None,
        description="Instruct supervisord to switch users to this UNIX user account before doing any meaningful processing. The user can only be switched if supervisord is started as the root user.",
    )
    directory: Optional[Path] = Field(
        default=None,
        description="When supervisord daemonizes, switch to this directory. This option can include the value %(here)s, which expands to the directory in which the supervisord configuration file was found.",
    )
    strip_ansi: Optional[bool] = Field(default=None, description="Strip all ANSI escape sequences from child log files.")
    environment: Optional[dict] = Field(
        default=None,
        description='A list of key/value pairs in the form KEY="val",KEY2="val2" that will be placed in the environment of all child processes. This does not change the environment of supervisord itself. This option can include the value %(here)s, which expands to the directory in which the supervisord configuration file was found. Values containing non-alphanumeric characters should be quoted (e.g. KEY="val:123",KEY2="val,456"). Otherwise, quoting the values is optional but recommended. To escape percent characters, simply use two. (e.g. URI="/first%%20name") Note that subprocesses will inherit the environment variables of the shell used to start supervisord except for the ones overridden here and within the program’s environment option. See Subprocess Environment.',
    )
    identifier: Optional[str] = Field(default=None, description="The identifier string for this supervisor process, used by the RPC interface.")

    @field_serializer("environment", when_used="json")
    def _dump_environment(self, v):
        if v:
            return ",".join(f"{k}={v}" for k, v in v.items())
        return None

    @field_validator("environment", mode="before")
    @classmethod
    def _load_environment(cls, v):
        if v:
            d = {}
            splits = v.split(",")
            for split in splits:
                split_splits = split.split("=")
                d[split_splits[0]] = split_splits[1]
            return d
        return v



from pydantic import ValidationError
from pytest import raises

from supervisor_pydantic import ProgramConfiguration, SupervisorConfiguration


def test_config_instantiation():
    with raises(ValidationError):
        c = SupervisorConfiguration()
    c = SupervisorConfiguration(program={"test": ProgramConfiguration(command="echo 'hello'")})
    assert c



from getpass import getuser
from pathlib import Path
from unittest.mock import patch

from supervisor_pydantic import load_config


def test_hydra_config():
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
    ):
        pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
        p1.return_value = str(pth)
        cfg = load_config("config", overrides=["+program=[sleep,echo]", "+rpcinterface=standard", "+inet_http_server=local"])
        assert (
            cfg.to_cfg().strip()
            == """[inet_http_server]
port=localhost:8000

[supervisord]
logfile={dir}/supervisord.log
pidfile={dir}/supervisord.pid
directory={dir}

[supervisorctl]

[program:sleep]
command=sleep 1000
directory={dir}/sleep

[program:echo]
command=echo
directory={dir}/echo

[rpcinterface:supervisor]
supervisor.rpcinterface_factory=supervisor.rpcinterface:make_main_rpcinterface""".format(dir=str(pth / f"supervisor-{getuser()}-sleep-echo"))
        )



from typing import List, Optional

from pydantic import Field, field_serializer, field_validator

from .base import EventType
from .program import ProgramConfiguration

__all__ = ("EventListenerConfiguration",)


class EventListenerConfiguration(ProgramConfiguration):
    def to_cfg(self, key: str) -> str:
        # Overload to require key
        return super().to_cfg(key=key).replace("[event_listener", "[eventlistener")

    buffer_size: Optional[int] = Field(
        default=None,
        description="The event listener pool’s event queue buffer size. When a listener pool’s event buffer is overflowed (as can happen when an event listener pool cannot keep up with all of the events sent to it), the oldest event in the buffer is discarded.",
    )
    events: Optional[List[EventType]] = Field(
        default=None,
        description="A comma-separated list of event type names that this listener is “interested” in receiving notifications for (see Event Types for a list of valid event type names).",
    )
    result_handler: Optional[str] = Field(
        default=None,
        description="A pkg_resources entry point string that resolves to a Python callable. The default value is supervisor.dispatchers:default_handler. Specifying an alternate result handler is a very uncommon thing to need to do, and as a result, how to create one is not documented.",
    )

    # TODO can't inherit stdout_capture_maxbytes
    @field_validator("stdout_capture_maxbytes")
    @classmethod
    def _event_listener_cant_use_stdout_capture_maxbytes(cls, v: str) -> str:
        if v:
            raise ValueError("eventlistener cannot use stdout_capture_maxbytes")
        return None

    @field_serializer("events", when_used="json")
    def _dump_events(self, v):
        if v:
            return ",".join(v)
        return None

    @field_validator("events", mode="before")
    @classmethod
    def _load_events(cls, v):
        if isinstance(v, str):
            return v.split(",")
        return v



from pydantic import ValidationError
from pytest import raises

from supervisor_pydantic import FcgiProgramConfiguration


def test_inst():
    with raises(ValidationError):
        FcgiProgramConfiguration()
    with raises(ValidationError):
        FcgiProgramConfiguration(socket="test")
    with raises(ValidationError):
        FcgiProgramConfiguration(command="test")
    FcgiProgramConfiguration(command="echo 'test'", socket="test")


def test_cfg():
    c = FcgiProgramConfiguration(command="echo 'test'", socket="test")
    assert c.to_cfg("name").strip() == "[fcgi-program:name]\ncommand=echo 'test'\nsocket=test"



from logging import getLogger
from pathlib import Path
from time import sleep
from typing import Callable, Optional, Union

from typer import Argument, Exit, Option, Typer
from typing_extensions import Annotated

from ..client import SupervisorRemoteXMLRPCClient
from ..config import SupervisorConvenienceConfiguration
from .common import SupervisorTaskStep

log = getLogger(__name__)

__all__ = (
    "write_supervisor_config",
    "start_supervisor",
    "start_programs",
    "check_programs",
    "stop_programs",
    "restart_programs",
    "stop_supervisor",
    "kill_supervisor",
    "remove_supervisor_config",
    "main",
)


def _check_exists(cfg: SupervisorConvenienceConfiguration) -> bool:
    log.info(f"Checking for working dir {cfg.working_dir} and config path {cfg.config_path}")
    if cfg.working_dir.exists() and cfg.config_path.exists():
        # its probably already been written
        log.info(f"Working dir {cfg.working_dir} and config path {cfg.config_path} exist")
        return True
    log.info(f"Working dir {cfg.working_dir} and/or config path {cfg.config_path} do not exist")
    return False


def _check_same(cfg: SupervisorConvenienceConfiguration) -> bool:
    log.info("Checking if config file matches")
    if _check_exists(cfg) and cfg.config_path.read_text().strip() == cfg.to_cfg().strip():
        # same file contents
        log.info(f"Config file {cfg.config_path} matches")
        return True
    elif not _check_exists(cfg):
        # no file exists, so we can write it
        log.info(f"Config file {cfg.config_path} does not exist, so we can write it")
        return True
    log.info(f"Config file {cfg.config_path} does not match")
    return False


def _check_running(cfg: SupervisorConvenienceConfiguration) -> bool:
    if _check_same(cfg):
        # check if running
        log.info("Checking if supervisor is running")
        if cfg.running():
            log.info("Supervisor is running")
            return True
        log.info("Supervisor is not running")
    log.info("Supervisor config file does not match, so we can't check if it's running")
    return False


def _wait_or_while(until: Callable, unless: Optional[Callable] = None, timeout: int = 5) -> bool:
    log.info(f"Waiting for {timeout} seconds")
    for _ in range(timeout):
        if until():
            log.info("`until` condition met")
            return True
        if unless and unless():
            log.info("`unless` condition met")
            return False
        sleep(1)
    log.info("Timed out after {timeout} seconds")
    return False


def _raise_or_exit(val: bool, exit: bool):
    if exit:
        raise Exit(int(not val))
    return val


def _load_or_pass(cfg: Union[str, SupervisorConvenienceConfiguration]) -> SupervisorConvenienceConfiguration:
    if isinstance(cfg, Path):
        cfg = SupervisorConvenienceConfiguration.model_validate_json(cfg.read_text())
    if isinstance(cfg, str):
        cfg = SupervisorConvenienceConfiguration.model_validate_json(cfg)
    if not isinstance(cfg, SupervisorConvenienceConfiguration):
        raise NotImplementedError
    return cfg


def write_supervisor_config(cfg_json: str, _exit: Annotated[bool, Argument(hidden=True)] = True):
    """Write a SupervisorConvenienceConfiguration JSON as a supervisor config file

    Args:
        cfg_json (str): JSON string of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    log.info(f"Loading JSON config: {cfg_json}")
    cfg_obj = _load_or_pass(cfg_json)

    if not _check_same(cfg_obj):
        log.critical("Configurations don't match while writing supervisor config. This may lead to zombie supervisors")

    log.info(f"Writing supervisor config to {cfg_obj.config_path}")
    cfg_obj._write_self()

    return _raise_or_exit(True, _exit)


def start_supervisor(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Start a supervisor instance using supervisord in background

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)

    if not _check_same(cfg_obj):
        log.critical("Configurations don't match while writing supervisor config. This may lead to zombie supervisors")

        # TODO check if "critical" things are different and restart
        # supervisor if necessary

        # Otherwise just write and reload
        log.info("Writing supervisor config to {cfg_obj.config_path}")
        cfg_obj._write_self()

        log.info("Reloading supervisor config")
        client = SupervisorRemoteXMLRPCClient(cfg=cfg_obj)
        client.reloadConfig()

    if _check_running(cfg_obj):
        log.info("Supervisor is already running")
        return _raise_or_exit(True, _exit)

    log.info("Starting supervisor")
    cfg_obj.start(daemon=True)
    running = _wait_or_while(until=lambda: cfg_obj.running(), timeout=cfg_obj.command_timeout)
    if not running:
        log.critical(f"Supervisor still not running {cfg_obj.command_timeout}s after start command!")
        return _raise_or_exit(False, _exit)
    log.info("Supervisor started")
    return _raise_or_exit(True, _exit)


def start_programs(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    restart: bool = False,
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Start all programs in the supervisor instance

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
        restart (bool, optional): if true, restart all programs. Defaults to False.
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)

    log.info(f"Initializing client with config: {cfg_obj}")
    client = SupervisorRemoteXMLRPCClient(cfg=cfg_obj)

    if restart:
        # No builtin restart, so stop everything then start again on the next line
        log.info("Stopping all processes before restarting")
        client.stopAllProcesses()

    log.info("Starting all processes")
    ret = client.startAllProcesses()
    log.info(ret)

    _wait_or_while(
        until=lambda: all(_.running() for _ in client.getAllProcessInfo()),
        unless=lambda: any(_.stopped() for _ in client.getAllProcessInfo()),
        timeout=cfg_obj.command_timeout,
    )
    all_ok = _wait_or_while(
        until=lambda: all(_.ok(ok_exitstatuses=cfg_obj.exitcodes) for _ in client.getAllProcessInfo()),
        unless=lambda: any(_.bad(ok_exitstatuses=cfg_obj.exitcodes) for _ in client.getAllProcessInfo()),
        timeout=cfg_obj.command_timeout,
    )
    if not all_ok:
        log.critical("Not all processes started successfully")
        for r in client.getAllProcessInfo():
            log.info(r.model_dump_json(exclude_unset=True))
        return _raise_or_exit(False, _exit)
    log.info("All processes started")
    return _raise_or_exit(True, _exit)


def check_programs(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    check_running: bool = False,
    check_done: bool = False,
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Check if programs are in a good state.

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
        check_running (bool, optional): if true, only return true if they're running
        check_done (bool, optional): if true, only return true if they're done (cleanly)
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)

    log.info(f"Initializing client with config: {cfg_obj}")
    client = SupervisorRemoteXMLRPCClient(cfg=cfg_obj)

    log.info("Checking all processes")
    ret = client.getAllProcessInfo()
    for r in ret:
        log.info(r.model_dump_json(exclude_unset=True))

    ok = False
    if check_running:
        if all(p.running() for p in ret):
            log.info("All processes running")
            ok = True
        else:
            log.warning("Not all processes running")
    elif check_done:
        if all(p.done(ok_exitstatuses=cfg_obj.exitcodes) for p in ret):
            log.info("All processes done")
            ok = True
        else:
            log.info("Not all processes done")
    else:
        if all(p.ok(ok_exitstatuses=cfg_obj.exitcodes) for p in ret):
            log.info("All processes ok")
            ok = True
        else:
            log.info("Not all processes ok")
    return _raise_or_exit(ok, _exit)


def stop_programs(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Stop all programs in the supervisor instance

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)

    log.info(f"Initializing client with config: {cfg_obj}")
    client = SupervisorRemoteXMLRPCClient(cfg=cfg_obj)

    log.info("Stopping all processes")
    ret = client.stopAllProcesses()
    log.info(ret)

    all_stopped = _wait_or_while(until=lambda: all(_.stopped() for _ in client.getAllProcessInfo()), timeout=cfg_obj.command_timeout)
    if not all_stopped:
        log.critical("Not all processes stopped successfully")
        for r in client.getAllProcessInfo():
            log.info(r.model_dump_json(exclude_unset=True))
        return _raise_or_exit(False, _exit)
    log.info("All processes stopped")
    return _raise_or_exit(True, _exit)


def restart_programs(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    force: bool = False,
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Restart all programs in the supervisor instance

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
        force (bool, optional): if true, force restart. Defaults to False.
    """
    if force:
        log.info("Force restarting all processes")
        if not stop_programs(cfg, False):
            log.warning("Could not stop programs")
            return _raise_or_exit(False, _exit)
    if not start_programs(cfg, False):
        log.warning("Could not start programs")
        return _raise_or_exit(False, _exit)
    return _raise_or_exit(True, _exit)


def stop_supervisor(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Stop the supervisor instance

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)

    log.info("Stopping supervisor")
    cfg_obj.stop()

    not_running = _wait_or_while(until=lambda: not cfg_obj.running(), timeout=cfg_obj.command_timeout)
    if not not_running:
        log.critical(f"Still running {cfg_obj.command_timeout}s after stop command!")
        return _raise_or_exit(False, _exit)
    return _raise_or_exit(True, _exit)


def kill_supervisor(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Kill the supervisor instance with os.kill

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    try:
        if not stop_programs(cfg, False):
            log.warning("could not stop programs")
    except ConnectionRefusedError:
        # supervisor already down, continue
        ...

    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)
    log.info("Killing supervisor")
    cfg_obj.kill()

    still_running = _wait_or_while(until=lambda: not cfg_obj.running(), timeout=cfg_obj.command_timeout)
    if still_running:
        log.critical(f"Still running {cfg_obj.command_timeout}s after kill command!")
        return _raise_or_exit(False, _exit)
    return _raise_or_exit(True, _exit)


def remove_supervisor_config(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Remove the supervisor config file and working directory

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)

    log.info("Removing supervisor config")
    still_running = stop_supervisor(cfg_obj, _exit=False)
    if still_running:
        log.critical("Supervisor still running after stop command!")
        still_running = kill_supervisor(cfg_obj, _exit=False)

    if still_running:
        log.critical("Supervisor still running after kill command!")
        return _raise_or_exit(False, _exit)

    log.info(f"Sleeping for {cfg_obj.command_timeout} seconds")
    sleep(cfg_obj.command_timeout)

    log.info("Removing supervisor config and folder")
    cfg_obj.rmdir()
    return _raise_or_exit(True, _exit)


def _add_to_typer(app, command: SupervisorTaskStep, foo):
    """Helper function to ensure correct command names"""
    app.command(command)(foo)


def main():
    app = Typer()
    _add_to_typer(app, "configure-supervisor", write_supervisor_config)
    _add_to_typer(app, "start-supervisor", start_supervisor)
    _add_to_typer(app, "start-programs", start_programs)
    _add_to_typer(app, "stop-programs", stop_programs)
    _add_to_typer(app, "check-programs", check_programs)
    _add_to_typer(app, "restart-programs", restart_programs)
    _add_to_typer(app, "stop-supervisor", stop_supervisor)
    _add_to_typer(app, "force-kill", kill_supervisor)
    _add_to_typer(app, "unconfigure-supervisor", remove_supervisor_config)
    app()



from supervisor_pydantic import SupervisordConfiguration


def test_inst():
    SupervisordConfiguration()


def test_cfg():
    c = SupervisordConfiguration(directory="/test")
    assert c.to_cfg().strip() == "[supervisord]\ndirectory=/test"



from supervisor_pydantic import UnixHttpServerConfiguration


def test_inst():
    UnixHttpServerConfiguration()
    UnixHttpServerConfiguration(file="/a/test/file")
    UnixHttpServerConfiguration(file="/a/test/file", username="test")
    UnixHttpServerConfiguration(file="/a/test/file", username="test", chmod="0777")


def test_cfg():
    c = UnixHttpServerConfiguration()
    assert c.to_cfg().strip() == """[unix_http_server]"""

    c = UnixHttpServerConfiguration(
        file="/a/test/file",
        chmod="0777",
        chown="test",
        username="test",
        password="testpw",
    )
    assert (
        c.to_cfg()
        == """[unix_http_server]
file=/a/test/file
chmod=0777
chown=test
username=test
password=testpw
"""
    )



from supervisor_pydantic import ProgramConfiguration, SupervisorConvenienceConfiguration


def test_inst():
    SupervisorConvenienceConfiguration(program={"test": ProgramConfiguration(command="echo test")})


def test_cfg_extra():
    c = SupervisorConvenienceConfiguration(
        port=7000,
        program={"test": ProgramConfiguration(command="echo test")},
        working_dir="/tmp/supervisor-runner-test",
    )
    assert (
        c.to_cfg().strip()
        == """[inet_http_server]
port=*:7000

[supervisord]
logfile=/tmp/supervisor-runner-test/supervisord.log
pidfile=/tmp/supervisor-runner-test/supervisord.pid
nodaemon=false
directory=/tmp/supervisor-runner-test
identifier=supervisor

[supervisorctl]
serverurl=http://localhost:7000/

[program:test]
command=echo test
autostart=false
startsecs=1
autorestart=false
exitcodes=0
stopsignal=TERM
stopwaitsecs=30
stopasgroup=true
killasgroup=true
stdout_logfile=/tmp/supervisor-runner-test/test/output.log
stderr_logfile=/tmp/supervisor-runner-test/test/error.log
directory=/tmp/supervisor-runner-test/test

[rpcinterface:supervisor]
supervisor.rpcinterface_factory=supervisor.rpcinterface:make_main_rpcinterface"""
    )



from helpers import get_calling_file


def test_get_calling_file_path():
    assert get_calling_file(0).endswith("supervisor_pydantic/utils.py")
    assert get_calling_file(1).endswith("helpers.py")
    assert get_calling_file(2).endswith("test_utils.py")



from pydantic import ValidationError
from pytest import raises

from supervisor_pydantic import EventListenerConfiguration


def test_inst():
    with raises(ValidationError):
        EventListenerConfiguration()
    with raises(ValidationError):
        EventListenerConfiguration(stdout_capture_maxbytes=10)
    EventListenerConfiguration(command="echo 'test'")


def test_cfg():
    c = EventListenerConfiguration(command="echo 'test'")
    assert c.to_cfg("name").strip() == "[eventlistener:name]\ncommand=echo 'test'"


def test_cfg_roundtrip():
    c = EventListenerConfiguration(command="echo 'test'")
    print(c.model_dump_json(exclude_unset=True))
    assert c.model_validate_json(c.model_dump_json(exclude_unset=True)) == c

    c = EventListenerConfiguration(command="echo 'test'", events=["PROCESS_STATE"])
    assert c.model_validate_json(c.model_dump_json()) == c



from .base import HostPort, LogLevel, Octal, OctalUmask, Signal, SupervisorLocation, UnixUserName, UnixUserNameOrGroup
from .convenience import SupervisorConvenienceConfiguration, load_convenience_config
from .eventlistener import EventListenerConfiguration
from .fcgiprogram import FcgiProgramConfiguration
from .group import GroupConfiguration
from .include import IncludeConfiguration
from .inet_http_server import InetHttpServerConfiguration
from .program import ProgramConfiguration
from .rpcinterface import RpcInterfaceConfiguration
from .supervisor import (
    SupervisorConfiguration,
    load_config,
)
from .supervisorctl import SupervisorctlConfiguration
from .supervisord import SupervisordConfiguration
from .unix_http_server import UnixHttpServerConfiguration



from typing import List, Optional

from pydantic import Field, field_serializer, field_validator

from .base import _BaseCfgModel

__all__ = ("GroupConfiguration",)


class GroupConfiguration(_BaseCfgModel):
    def to_cfg(self, key: str) -> str:
        # Overload to require key
        return super().to_cfg(key=key)

    programs: List[str] = Field(description="A comma-separated list of program names. The programs which are listed become members of the group.")
    priority: Optional[int] = Field(default=None, description="A priority number analogous to a [program:x] priority value assigned to the group.")

    @field_serializer("programs", when_used="json")
    def _dump_programs(self, v):
        if v:
            return ",".join(v)
        return None

    @field_validator("programs", mode="before")
    @classmethod
    def _load_programs(cls, v):
        if isinstance(v, str):
            return v.split(",")
        return v



from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional
from xmlrpc.client import Fault, ServerProxy

from pydantic import BaseModel

from ..config import SupervisorConvenienceConfiguration

__all__ = ("SupervisorRemoteXMLRPCClient", "ProcessState", "SupervisorState", "SupervisorMethodResult", "ProcessInfo")


class ProcessState(Enum):
    STOPPED = 0
    STARTING = 10
    RUNNING = 20
    BACKOFF = 30
    STOPPING = 40
    EXITED = 100
    FATAL = 200
    UNKNOWN = 1000

    @classmethod
    def _missing_(cls, code):
        if isinstance(code, str):
            return getattr(cls, code)
        if code not in (0, 10, 20, 30, 40, 100, 200):
            return super().__init__(1000)
        raise ValueError(code)


class SupervisorState(Enum):
    FATAL = 2
    RUNNING = 1
    RESTARTING = 0
    SHUTDOWN = -1

    @classmethod
    def _missing_(cls, code):
        if isinstance(code, str):
            return getattr(cls, code)
        raise ValueError(code)


class SupervisorMethodResult(Enum):
    # duplicated from https://github.com/Supervisor/supervisor/blob/29eeb9dd55c55da2e83c5497d01f3a859998ecf9/supervisor/xmlrpc.py
    UNKNOWN_METHOD = 1
    INCORRECT_PARAMETERS = 2
    BAD_ARGUMENTS = 3
    SIGNATURE_UNSUPPORTED = 4
    SHUTDOWN_STATE = 6
    BAD_NAME = 10
    BAD_SIGNAL = 11
    NO_FILE = 20
    NOT_EXECUTABLE = 21
    FAILED = 30
    ABNORMAL_TERMINATION = 40
    SPAWN_ERROR = 50
    ALREADY_STARTED = 60
    NOT_RUNNING = 70
    SUCCESS = 80
    ALREADY_ADDED = 90
    STILL_RUNNING = 91
    CANT_REREAD = 92


class ProcessInfo(BaseModel):
    name: str
    group: str
    state: ProcessState
    description: str
    start: datetime
    stop: datetime
    now: datetime
    spawner: str = ""
    exitstatus: int
    logfile: str
    stdout_logfile: str
    stderr_logfile: str
    pid: int

    def running(self):
        return self.state in (ProcessState.RUNNING, ProcessState.STOPPING)

    def stopped(self):
        return self.state in (ProcessState.STOPPED, ProcessState.EXITED, ProcessState.FATAL)

    def done(self, ok_exitstatuses=None):
        ok_exitstatuses = ok_exitstatuses or (0,)
        return self.state in (ProcessState.STOPPED,) or (self.state == ProcessState.EXITED and self.exitstatus in ok_exitstatuses)

    def ok(self, ok_exitstatuses=None):
        ok_exitstatuses = ok_exitstatuses or (0,)
        return self.state in (
            # ProcessState.STARTING,
            ProcessState.RUNNING,
            ProcessState.STOPPING,
            ProcessState.STOPPED,
        ) or (self.state == ProcessState.EXITED and self.exitstatus in ok_exitstatuses)

    def bad(self, ok_exitstatuses=None):
        ok_exitstatuses = ok_exitstatuses or (0,)
        return self.state in (ProcessState.FATAL, ProcessState.UNKNOWN) or (
            self.state == ProcessState.EXITED and self.exitstatus not in ok_exitstatuses
        )


class SupervisorRemoteXMLRPCClient(object):
    """A light wrapper over the supervisor xmlrpc api: http://supervisord.org/api.html"""

    def __init__(self, cfg: SupervisorConvenienceConfiguration):
        self._cfg = cfg
        self._host = cfg.host
        self._port = int(cfg.port.split(":")[-1])
        self._protocol = cfg.protocol
        self._rpcpath = "/" + cfg.rpcpath if not cfg.rpcpath.startswith("/") else cfg.rpcpath
        self._rpcurl = self._build_rpcurl(username=cfg.username, password=cfg.password)
        self._client = ServerProxy(self._rpcurl)

    def _build_rpcurl(self, username: Optional[str], password: Optional[str]) -> str:
        # Forces http or https based on port, otherwise resolves to given protocol
        protocol = {80: "http", 443: "https"}.get(self._port, self._protocol)
        port = "" if self._port in {80, 443} else f":{self._port}"
        authentication = f"{username}:{password.get_secret_value()}@" if username and password else ""
        return f"{protocol}://{authentication}{self._host}{port}{self._rpcpath}"

    #######################
    # supervisord methods #
    #######################
    def getAllProcessInfo(self) -> List[ProcessInfo]:
        return [ProcessInfo(**_) for _ in self._client.supervisor.getAllProcessInfo()]

    def getState(self) -> SupervisorState:
        return SupervisorState(self._client.supervisor.getState()["statecode"])

    # def readLog(self):
    #     return self._client.supervisor.readLog(0, 0)

    def restart(self) -> SupervisorState:
        self._client.supervisor.restart()
        return self.getState()

    def shutdown(self) -> SupervisorState:
        self._client.supervisor.shutdown()
        return self.getState()

    ###################
    # process methods #
    ###################
    def getProcessInfo(self, name: str) -> ProcessInfo:
        if name not in self._cfg.program:
            raise RuntimeError(f"Unknown process: {name}")
        return self._getProcessInfoInternal(name)

    def _getProcessInfoInternal(self, name: str) -> ProcessInfo:
        return ProcessInfo(**self._client.supervisor.getProcessInfo(name))

    def readProcessLog(self, name: str):
        if name not in self._cfg.program:
            raise RuntimeError(f"Unknown process: {name}")
        return self._client.supervisor.readProcessLog(name, 0, 0)

    def readProcessStderrLog(self, name: str):
        if name not in self._cfg.program:
            raise RuntimeError(f"Unknown process: {name}")

        return self._client.supervisor.readProcessStderrLog()

    def readProcessStdoutLog(self, name: str):
        return self._client.supervisor.readProcessStdoutLog()

    def startAllProcesses(self) -> Dict[str, ProcessInfo]:
        # start all
        self._client.supervisor.startAllProcesses()
        return {name: self.getProcessInfo(name) for name in self._cfg.program}

    def startProcess(self, name: str) -> ProcessInfo:
        if name not in self._cfg.program:
            raise RuntimeError(f"Unknown process: {name}")
        try:
            if self._client.supervisor.startProcess(name):
                return self.getProcessInfo(name)
        except Fault as f:
            if f.faultCode == SupervisorMethodResult.ALREADY_STARTED.value:
                return self.getProcessInfo(name)
            if f.faultCode == SupervisorMethodResult.SPAWN_ERROR.value:
                return self.getProcessInfo(name)
            raise f
        return self.getProcessInfo(name)

    def stopAllProcesses(self) -> Dict[str, ProcessInfo]:
        # start all
        self._client.supervisor.stopAllProcesses()
        return {name: self.getProcessInfo(name) for name in self._cfg.program}

    def stopProcess(self, name: str) -> ProcessInfo:
        if name not in self._cfg.program:
            raise RuntimeError(f"Unknown process: {name}")
        return self._stopProcessInternal(name)

    def _stopProcessInternal(self, name: str) -> ProcessInfo:
        try:
            if self._client.supervisor.stopProcess(name):
                return self._getProcessInfoInternal(name)
        except Fault as f:
            if f.faultCode == SupervisorMethodResult.NOT_RUNNING.value:
                return self._getProcessInfoInternal(name)
            raise f
        return self._getProcessInfoInternal(name)

    def reloadConfig(self, start_new: bool = False) -> SupervisorState:
        added, changed, removed = self._client.supervisor.reloadConfig()[0]
        proc_infos = []
        for name in removed:
            proc_infos.append(self._stopProcessInternal(name))
        for name in changed:
            self._stopProcessInternal(name)
            proc_infos.append(self.startProcess(name))
        # Don't need to start as we'll do this separately
        for name in added:
            self._client.supervisor.addProcessGroup(name)
            if start_new:
                proc_infos.append(self.startProcess(name))
        return proc_infos

    # def signalAllProcesses(self, signal):
    #     return self._client.supervisor.signalAllProcesses()

    def signalProcess(self, name: str, signal):
        if name not in self._cfg.program:
            raise RuntimeError(f"Unknown process: {name}")
        return self._client.supervisor.signalProcess()



from typing import Optional

from pydantic import Field, SecretStr, field_serializer

from .base import HostPort, UnixUserName, _BaseCfgModel

__all__ = ("InetHttpServerConfiguration",)


class InetHttpServerConfiguration(_BaseCfgModel):
    def to_cfg(self) -> str:
        ret = "[inet_http_server]"
        if self.port:
            ret += f"\nport={self.port}"
        if self.username:
            ret += f"\nusername={self.username}"
        if self.password:
            ret += f"\npassword={self.password.get_secret_value()}"
        return ret.strip() + "\n"

    port: Optional[HostPort] = Field(
        default=None,
        description="A TCP host:port value or (e.g. 127.0.0.1:9001) on which supervisor will listen for HTTP/XML-RPC requests. supervisorctl will use XML-RPC to communicate with supervisord over this port. To listen on all interfaces in the machine, use :9001 or *:9001. Please read the security warning above.",
    )
    username: Optional[UnixUserName] = Field(default=None, description="The username required for authentication to this HTTP server.")
    password: Optional[SecretStr] = Field(
        default=None,
        description="he password required for authentication to this HTTP server. This can be a cleartext password, or can be specified as a SHA-1 hash if prefixed by the string {SHA}. For example, {SHA}82ab876d1387bfafe46cc1c8a2ef074eae50cb1d is the SHA-stored version of the password “thepassword”. Note that hashed password must be in hex format.",
    )

    @field_serializer("password", when_used="json")
    def _dump_password(self, v):
        return v.get_secret_value() if v else None



from pathlib import Path
from typing import Optional

from pydantic import Field, SecretStr, field_serializer

from .base import Octal, UnixUserName, UnixUserNameOrGroup, _BaseCfgModel

__all__ = ("UnixHttpServerConfiguration",)


class UnixHttpServerConfiguration(_BaseCfgModel):
    file: Optional[Path] = Field(
        default=None,
        description="A path to a UNIX domain socket on which supervisor will listen for HTTP/XML-RPC requests. supervisorctl uses XML-RPC to communicate with supervisord over this port. This option can include the value %(here)s, which expands to the directory in which the supervisord configuration file was found.",
    )
    chmod: Optional[Octal] = Field(
        default=None,
        description="Change the UNIX permission mode bits of the UNIX domain socket to this value at startup.",
    )
    chown: Optional[UnixUserNameOrGroup] = Field(
        default=None,
        description="Change the user and group of the socket file to this value. May be a UNIX username (e.g. chrism) or a UNIX username and group separated by a colon (e.g. chrism:wheel).",
    )
    username: Optional[UnixUserName] = Field(default=None, description="The username required for authentication to this HTTP server.")
    password: Optional[SecretStr] = Field(
        default=None,
        description="The password required for authentication to this HTTP server. This can be a cleartext password, or can be specified as a SHA-1 hash if prefixed by the string {SHA}. For example, {SHA}82ab876d1387bfafe46cc1c8a2ef074eae50cb1d is the SHA-stored version of the password “thepassword”. Note that hashed password must be in hex format.",
    )

    @field_serializer("password", when_used="json")
    def _dump_password(self, v):
        return v.get_secret_value() if v else None



from typing import Optional

from pydantic import Field

from .base import Octal, UnixUserNameOrGroup
from .program import ProgramConfiguration

__all__ = ("FcgiProgramConfiguration",)


class FcgiProgramConfiguration(ProgramConfiguration):
    def to_cfg(self, key: str) -> str:
        # Overload to require key
        return super().to_cfg(key=key).replace("[fcgi_program", "[fcgi-program")

    socket: str = Field(
        description="The FastCGI socket for this program, either TCP or UNIX domain socket. For TCP sockets, use this format: tcp://localhost:9002. For UNIX domain sockets, use unix:///absolute/path/to/file.sock. String expressions are evaluated against a dictionary containing the keys “program_name” and “here” (the directory of the supervisord config file)."
    )
    socket_backlog: Optional[str] = Field(default=None, description="Sets socket listen(2) backlog.")
    socket_owner: Optional[UnixUserNameOrGroup] = Field(
        default=None,
        description="For UNIX domain sockets, this parameter can be used to specify the user and group for the FastCGI socket. May be a UNIX username (e.g. chrism) or a UNIX username and group separated by a colon (e.g. chrism:wheel).",
    )
    socket_mode: Optional[Octal] = Field(
        default=None, description="For UNIX domain sockets, this parameter can be used to specify the permission mode."
    )



from supervisor_pydantic import SupervisorctlConfiguration


def test_inst():
    SupervisorctlConfiguration()


def test_cfg():
    c = SupervisorctlConfiguration(username="test", password="testpw")
    assert c.to_cfg().strip() == "[supervisorctl]\nusername=test\npassword=testpw"



import xmlrpc
from time import sleep

import pytest

from supervisor_pydantic import SupervisorConvenienceConfiguration, SupervisorRemoteXMLRPCClient
from supervisor_pydantic.client.xmlrpc import ProcessState


def _assert_client_actions(client: SupervisorRemoteXMLRPCClient, name: str = "test"):
    assert client.getProcessInfo(name).state == ProcessState.STOPPED
    sleep(1)
    assert client.startAllProcesses()[name].state == ProcessState.RUNNING
    sleep(1)
    assert client.getProcessInfo(name).state == ProcessState.EXITED
    assert client.startProcess(name).state == ProcessState.RUNNING
    assert client.stopProcess(name).state == ProcessState.STOPPED
    assert client.startProcess(name).state == ProcessState.RUNNING
    assert client.stopAllProcesses()[name].state == ProcessState.STOPPED


def test_supervisor_client(supervisor_instance: SupervisorConvenienceConfiguration):
    client = SupervisorRemoteXMLRPCClient(supervisor_instance)
    _assert_client_actions(client=client)


def test_supervisor_client_changes(supervisor_instance: SupervisorConvenienceConfiguration):
    client = SupervisorRemoteXMLRPCClient(supervisor_instance)
    _assert_client_actions(client=client)
    test_program = supervisor_instance.program.pop("test")
    supervisor_instance.program["test2"] = test_program
    supervisor_instance._write_self()
    client = SupervisorRemoteXMLRPCClient(supervisor_instance)
    client.reloadConfig()
    print(supervisor_instance)
    _assert_client_actions(client=client, name="test2")


def test_permissioned_supervisor_client_rejected(permissioned_supervisor_instance: SupervisorConvenienceConfiguration):
    permissioned_supervisor_instance.username = "bad-username"
    client = SupervisorRemoteXMLRPCClient(permissioned_supervisor_instance)
    with pytest.raises(xmlrpc.client.ProtocolError):
        client.getProcessInfo("test")


def test_permissioned_supervisor_client(permissioned_supervisor_instance: SupervisorConvenienceConfiguration):
    permissioned_supervisor_instance.username = "user1"
    client = SupervisorRemoteXMLRPCClient(permissioned_supervisor_instance)
    _assert_client_actions(client=client)



from typing import List

from pydantic import Field, field_serializer, field_validator

from .base import _BaseCfgModel

__all__ = ("IncludeConfiguration",)


class IncludeConfiguration(_BaseCfgModel):
    files: List[str] = Field(
        description="A space-separated sequence of file globs. Each file glob may be absolute or relative. If the file glob is relative, it is considered relative to the location of the configuration file which includes it. A “glob” is a file pattern which matches a specified pattern according to the rules used by the Unix shell. No tilde expansion is done, but *, ?, and character ranges expressed with [] will be correctly matched. The string expression is evaluated against a dictionary that includes host_node_name and here (the directory of the supervisord config file). Recursive includes from included files are not supported."
    )

    @field_serializer("files", when_used="json")
    def _dump_files(self, v):
        if v:
            return " ".join(v)
        return None

    @field_validator("files", mode="before")
    @classmethod
    def _load_files(cls, v):
        if isinstance(v, str):
            return v.split(" ")
        return v
